; -*- Scheme -*-
;
; $Id: string48.scm,v 1.1 1998/03/16 07:59:57 foner Exp $

;------------

(require 'char-set?)
(require 'char-set:whitespace)
(require 'string->char-set)
(require 'string:trim-left-with-chars)
(require 'string:trim-left-with-char)
(require 'string:trim-left-by-proc)

;------------

;+doc
; procedure: string:trim-left
; arguments: string chars
; signature: string x char-set -> string
; arguments: string char
; signature: string x char -> string
; arguments: string string
; signature: string x string -> string
; arguments: string proc
; signature: string x (char -> bool) -> string
; arguments: string
; signature: string -> string
;
; Return a fresh copy of STRING with certain characters removed from
; the left end.  Various methods of trimming are supported.  These
; are trimming using :-
; 
;  * a char-set to define all the characters to be removed
;  * a single character defining the character to be removed
;  * a string which is converted into a char-set and then used.
;  * a procedure which should return #t for any character to be removed.
;  * whitespace if no method of trimming is supplied.
;
; Examples:
;
; Three different ways of removing whitespace :-
;
; > (string:trim-left "    Betelgeuse")
; "Betelgeuse"
; > (string:trim-left "    Betelgeuse" char-whitespace?)
; "Betelgeuse"
; > (string:trim-left "    Betelgeuse" char-set:whitespace)
; "Betelgeuse"
;-doc

(define string:trim-left
  (lambda (string . opt)
    (let ((m (if (null? opt) char-set:whitespace (car opt))))
      (cond
       ((char-set? m) (string:trim-left-with-chars string m))
       ((string? m) (string:trim-left-with-chars string (string->char-set m)))
       ((char? m) (string:trim-left-with-char string m))
       ((procedure? m) (string:trim-left-by-proc string m))
       (else (error 'string:trim-left "invalid trim method"))))))

; eof
